@@ -3,7 +3,6 @@
 #include <err.h>
 #include <errno.h>
 #include <inttypes.h>
-#include <libgen.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -13,7 +12,6 @@
 #include <git2.h>
 
 #include "compat.h"
-#include "config.h"
 
 static git_repository *repo;
 
@@ -21,18 +19,27 @@ static const char *relpath = "";
 static const char *repodir;
 
 static char description[255] = "Repositories";
-static char name[255];
+static char *name = "";
 static char owner[255];
 
-FILE *
-efopen(const char *name, const char *flags)
+#ifndef USE_PLEDGE
+int
+pledge(const char *promises, const char *paths[])
 {
-	FILE *fp;
+        return 0;
+}
+#endif
 
-	if (!(fp = fopen(name, flags)))
-		err(1, "fopen");
+void
+joinpath(char *buf, size_t bufsiz, const char *path, const char *path2)
+{
+	int r;
 
-	return fp;
+	r = snprintf(buf, bufsiz, "%s%s%s",
+		path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
+	if (r == -1 || (size_t)r >= bufsiz)
+		errx(1, "path truncated: '%s%s%s'",
+			path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
 }
 
 /* Escape characters below as HTML 2.0 / XML 1.0. */
@@ -53,47 +60,21 @@ xmlencode(FILE *fp, const char *s, size_t len)
 	}
 }
 
-/* Some implementations of basename(3) return a pointer to a static
- * internal buffer (OpenBSD). Others modify the contents of `path` (POSIX).
- * This is a wrapper function that is compatible with both versions.
- * The program will error out if basename(3) failed, this can only happen
- * with the OpenBSD version. */
-char *
-xbasename(const char *path)
-{
-	char *p, *b;
-
-	if (!(p = strdup(path)))
-		err(1, "strdup");
-	if (!(b = basename(p)))
-		err(1, "basename");
-	if (!(b = strdup(b)))
-		err(1, "strdup");
-	free(p);
-
-	return b;
-}
-
 void
-printtimeformat(FILE *fp, const git_time *intime, const char *fmt)
+printtimeshort(FILE *fp, const git_time *intime)
 {
 	struct tm *intm;
 	time_t t;
 	char out[32];
 
-	t = (time_t) intime->time + (intime->offset * 60);
-	intm = gmtime(&t);
-	strftime(out, sizeof(out), fmt, intm);
+	t = (time_t)intime->time;
+	if (!(intm = gmtime(&t)))
+		return;
+	strftime(out, sizeof(out), "%Y-%m-%d %H:%M", intm);
 	fputs(out, fp);
 }
 
 void
-printtimeshort(FILE *fp, const git_time *intime)
-{
-	printtimeformat(fp, intime, "%Y-%m-%d %H:%M");
-}
-
-int
 writeheader(FILE *fp)
 {
 	fputs("<!DOCTYPE html>\n"
@@ -111,24 +92,22 @@ writeheader(FILE *fp)
 	      "<table id=\"index\"><thead>\n"
 	      "<tr><td>Name</td><td>Description</td><td>Owner</td><td>Last commit</td></tr>"
 	      "</thead><tbody>\n", fp);
-
-	return 0;
 }
 
-int
+void
 writefooter(FILE *fp)
 {
-	return !fputs("</tbody>\n</table>\n</div>\n</body>\n</html>\n", fp);
+	fputs("</tbody>\n</table>\n</div>\n</body>\n</html>\n", fp);
 }
 
 int
 writelog(FILE *fp)
 {
-	char *stripped_name = NULL, *p;
 	git_commit *commit = NULL;
 	const git_signature *author;
 	git_revwalk *w = NULL;
 	git_oid id;
+	char *stripped_name = NULL, *p;
 	int ret = 0;
 
 	git_revwalk_new(&w, repo);
@@ -177,8 +156,11 @@ main(int argc, char *argv[])
 {
 	const git_error *e = NULL;
 	FILE *fp;
-	char path[PATH_MAX], *p;
-	int i, r, ret = 0;
+	char path[PATH_MAX], repodirabs[PATH_MAX + 1];
+	int i, ret = 0;
+
+	if (pledge("stdio rpath", NULL) == -1)
+		err(1, "pledge");
 
 	if (argc < 2) {
 		fprintf(stderr, "%s [repodir...]\n", argv[0]);
@@ -190,6 +172,8 @@ main(int argc, char *argv[])
 
 	for (i = 1; i < argc; i++) {
 		repodir = argv[i];
+		if (!realpath(repodir, repodirabs))
+			err(1, "realpath");
 
 		if (git_repository_open_ext(&repo, repodir,
 		    GIT_REPOSITORY_OPEN_NO_SEARCH, NULL)) {
@@ -199,26 +183,19 @@ main(int argc, char *argv[])
 			continue;
 		}
 
-		/* use directory name as name, truncation of name is no problem. */
-		p = xbasename(repodir);
-		snprintf(name, sizeof(name), "%s", p);
-		free(p);
+		/* use directory name as name */
+		if ((name = strrchr(repodirabs, '/')))
+			name++;
+		else
+			name = "";
 
 		/* read description or .git/description */
-		description[0] = '\0';
-		r = snprintf(path, sizeof(path), "%s%s%s",
-			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: '%s%s%s'",
-			        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
+		joinpath(path, sizeof(path), repodir, "description");
 		if (!(fp = fopen(path, "r"))) {
-			r = snprintf(path, sizeof(path), "%s%s%s",
-				repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
-			if (r == -1 || (size_t)r >= sizeof(path))
-				errx(1, "path truncated: '%s%s%s'",
-				        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
+			joinpath(path, sizeof(path), repodir, ".git/description");
 			fp = fopen(path, "r");
 		}
+		description[0] = '\0';
 		if (fp) {
 			if (!fgets(description, sizeof(description), fp))
 				description[0] = '\0';
@@ -226,20 +203,12 @@ main(int argc, char *argv[])
 		}
 
 		/* read owner or .git/owner */
-		owner[0] = '\0';
-		r = snprintf(path, sizeof(path), "%s%s%s",
-			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "owner");
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: '%s%s%s'",
-			        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "owner");
+		joinpath(path, sizeof(path), repodir, "owner");
 		if (!(fp = fopen(path, "r"))) {
-			r = snprintf(path, sizeof(path), "%s%s%s",
-				repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/owner");
-			if (r == -1 || (size_t)r >= sizeof(path))
-				errx(1, "path truncated: '%s%s%s'",
-				        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/owner");
+			joinpath(path, sizeof(path), repodir, ".git/owner");
 			fp = fopen(path, "r");
 		}
+		owner[0] = '\0';
 		if (fp) {
 			if (!fgets(owner, sizeof(owner), fp))
 				owner[0] = '\0';
