@@ -13,7 +13,13 @@
 #include <git2.h>
 
 #include "compat.h"
-#include "config.h"
+
+struct deltainfo {
+	git_patch *patch;
+
+	size_t addcount;
+	size_t delcount;
+};
 
 struct commitinfo {
 	const git_oid *id;
@@ -22,39 +28,155 @@ struct commitinfo {
 	char parentoid[GIT_OID_HEXSZ + 1];
 
 	const git_signature *author;
+	const git_signature *committer;
 	const char          *summary;
 	const char          *msg;
 
-	git_diff_stats *stats;
-	git_diff       *diff;
-	git_commit     *commit;
-	git_commit     *parent;
-	git_tree       *commit_tree;
-	git_tree       *parent_tree;
+	git_diff   *diff;
+	git_commit *commit;
+	git_commit *parent;
+	git_tree   *commit_tree;
+	git_tree   *parent_tree;
 
 	size_t addcount;
 	size_t delcount;
 	size_t filecount;
+
+	struct deltainfo **deltas;
+	size_t ndeltas;
 };
 
+/* summary length (bytes) in the log */
+static const unsigned summarylen = 70;
+/* display line count or file size in file tree index */
+static const int showlinecount = 1;
+
 static git_repository *repo;
 
 static const char *relpath = "";
 static const char *repodir;
 
-static char *name;
-static char *stripped_name;
+static char *name = "";
+static char *strippedname;
 static char description[255];
 static char cloneurl[1024];
-static int hasreadme, haslicense;
+static int haslicense, hasreadme, hassubmodules;
+
+/* cache */
+static git_oid lastoid;
+static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + nul byte */
+static FILE *rcachefp, *wcachefp;
+static const char *cachefile;
+
+#ifndef USE_PLEDGE
+int
+pledge(const char *promises, const char *paths[])
+{
+	return 0;
+}
+#endif
+
+void
+joinpath(char *buf, size_t bufsiz, const char *path, const char *path2)
+{
+	int r;
+
+	r = snprintf(buf, bufsiz, "%s%s%s",
+		path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
+	if (r == -1 || (size_t)r >= bufsiz)
+		errx(1, "path truncated: '%s%s%s'",
+			path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
+}
+
+void
+deltainfo_free(struct deltainfo *di)
+{
+	if (!di)
+		return;
+	git_patch_free(di->patch);
+	di->patch = NULL;
+	free(di);
+}
+
+int
+commitinfo_getstats(struct commitinfo *ci)
+{
+	struct deltainfo *di;
+	const git_diff_delta *delta;
+	const git_diff_hunk *hunk;
+	const git_diff_line *line;
+	git_patch *patch = NULL;
+	size_t ndeltas, nhunks, nhunklines;
+	size_t i, j, k;
+
+	ndeltas = git_diff_num_deltas(ci->diff);
+	if (ndeltas && !(ci->deltas = calloc(ndeltas, sizeof(struct deltainfo *))))
+		err(1, "calloc");
+
+	for (i = 0; i < ndeltas; i++) {
+		if (!(di = calloc(1, sizeof(struct deltainfo))))
+			err(1, "calloc");
+		if (git_patch_from_diff(&patch, ci->diff, i)) {
+			git_patch_free(patch);
+			goto err;
+		}
+		di->patch = patch;
+		ci->deltas[i] = di;
+
+		delta = git_patch_get_delta(patch);
+
+		/* skip stats for binary data */
+		if (delta->flags & GIT_DIFF_FLAG_BINARY)
+			continue;
+
+		nhunks = git_patch_num_hunks(patch);
+		for (j = 0; j < nhunks; j++) {
+			if (git_patch_get_hunk(&hunk, &nhunklines, patch, j))
+				break;
+			for (k = 0; ; k++) {
+				if (git_patch_get_line_in_hunk(&line, patch, j, k))
+					break;
+				if (line->old_lineno == -1) {
+					di->addcount++;
+					ci->addcount++;
+				} else if (line->new_lineno == -1) {
+					di->delcount++;
+					ci->delcount++;
+				}
+			}
+		}
+	}
+	ci->ndeltas = i;
+	ci->filecount = i;
+
+	return 0;
+
+err:
+	if (ci->deltas)
+		for (i = 0; i < ci->ndeltas; i++)
+			deltainfo_free(ci->deltas[i]);
+	free(ci->deltas);
+	ci->deltas = NULL;
+	ci->ndeltas = 0;
+	ci->addcount = 0;
+	ci->delcount = 0;
+	ci->filecount = 0;
+
+	return -1;
+}
 
 void
 commitinfo_free(struct commitinfo *ci)
 {
+	size_t i;
+
 	if (!ci)
 		return;
-
-	git_diff_stats_free(ci->stats);
+	if (ci->deltas)
+		for (i = 0; i < ci->ndeltas; i++)
+			deltainfo_free(ci->deltas[i]);
+	free(ci->deltas);
+	ci->deltas = NULL;
 	git_diff_free(ci->diff);
 	git_tree_free(ci->commit_tree);
 	git_tree_free(ci->parent_tree);
@@ -66,43 +188,38 @@ commitinfo_getbyoid(const git_oid *id)
 {
 	struct commitinfo *ci;
 	git_diff_options opts;
-	int error;
 
 	if (!(ci = calloc(1, sizeof(struct commitinfo))))
 		err(1, "calloc");
 
-	ci->id = id;
 	if (git_commit_lookup(&(ci->commit), repo, id))
 		goto err;
+	ci->id = id;
 
 	git_oid_tostr(ci->oid, sizeof(ci->oid), git_commit_id(ci->commit));
 	git_oid_tostr(ci->parentoid, sizeof(ci->parentoid), git_commit_parent_id(ci->commit, 0));
 
 	ci->author = git_commit_author(ci->commit);
+	ci->committer = git_commit_committer(ci->commit);
 	ci->summary = git_commit_summary(ci->commit);
 	ci->msg = git_commit_message(ci->commit);
 
-	if ((error = git_commit_tree(&(ci->commit_tree), ci->commit)))
+	if (git_tree_lookup(&(ci->commit_tree), repo, git_commit_tree_id(ci->commit)))
 		goto err;
-	if (!(error = git_commit_parent(&(ci->parent), ci->commit, 0))) {
-		if ((error = git_commit_tree(&(ci->parent_tree), ci->parent)))
-			goto err;
-	} else {
-		ci->parent = NULL;
-		ci->parent_tree = NULL;
+	if (!git_commit_parent(&(ci->parent), ci->commit, 0)) {
+		if (git_tree_lookup(&(ci->parent_tree), repo, git_commit_tree_id(ci->parent))) {
+			ci->parent = NULL;
+			ci->parent_tree = NULL;
+		}
 	}
 
 	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
 	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH;
-	if ((error = git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts)))
+	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
 		goto err;
-	if (git_diff_get_stats(&(ci->stats), ci->diff))
+	if (commitinfo_getstats(ci) == -1)
 		goto err;
 
-	ci->addcount = git_diff_stats_insertions(ci->stats);
-	ci->delcount = git_diff_stats_deletions(ci->stats);
-	ci->filecount = git_diff_stats_files_changed(ci->stats);
-
 	return ci;
 
 err:
@@ -141,48 +258,6 @@ xmlencode(FILE *fp, const char *s, size_t len)
 	}
 }
 
-/* Some implementations of dirname(3) return a pointer to a static
- * internal buffer (OpenBSD). Others modify the contents of `path` (POSIX).
- * This is a wrapper function that is compatible with both versions.
- * The program will error out if dirname(3) failed, this can only happen
- * with the OpenBSD version. */
-char *
-xdirname(const char *path)
-{
-	char *p, *b;
-
-	if (!(p = strdup(path)))
-		err(1, "strdup");
-	if (!(b = dirname(p)))
-		err(1, "dirname");
-	if (!(b = strdup(b)))
-		err(1, "strdup");
-	free(p);
-
-	return b;
-}
-
-/* Some implementations of basename(3) return a pointer to a static
- * internal buffer (OpenBSD). Others modify the contents of `path` (POSIX).
- * This is a wrapper function that is compatible with both versions.
- * The program will error out if basename(3) failed, this can only happen
- * with the OpenBSD version. */
-char *
-xbasename(const char *path)
-{
-	char *p, *b;
-
-	if (!(p = strdup(path)))
-		err(1, "strdup");
-	if (!(b = basename(p)))
-		err(1, "basename");
-	if (!(b = strdup(b)))
-		err(1, "strdup");
-	free(p);
-
-	return b;
-}
-
 int
 mkdirp(const char *path)
 {
@@ -204,37 +279,53 @@ mkdirp(const char *path)
 }
 
 void
-printtimeformat(FILE *fp, const git_time *intime, const char *fmt)
+printtimez(FILE *fp, const git_time *intime)
 {
 	struct tm *intm;
 	time_t t;
 	char out[32];
 
-	t = (time_t) intime->time + (intime->offset * 60);
-	intm = gmtime(&t);
-	strftime(out, sizeof(out), fmt, intm);
+	t = (time_t)intime->time;
+	if (!(intm = gmtime(&t)))
+		return;
+	strftime(out, sizeof(out), "%Y-%m-%dT%H:%M:%SZ", intm);
 	fputs(out, fp);
 }
 
-void
-printtimez(FILE *fp, const git_time *intime)
-{
-	printtimeformat(fp, intime, "%Y-%m-%dT%H:%M:%SZ");
-}
-
 void
 printtime(FILE *fp, const git_time *intime)
 {
-	printtimeformat(fp, intime, "%a %b %e %T %Y");
+	struct tm *intm;
+	time_t t;
+	char out[32];
+
+	t = (time_t)intime->time + (intime->offset * 60);
+	if (!(intm = gmtime(&t)))
+		return;
+	strftime(out, sizeof(out), "%a %b %e %H:%M:%S", intm);
+	if (intime->offset < 0)
+		fprintf(fp, "%s -%02d%02d", out,
+		            -(intime->offset) / 60, -(intime->offset) % 60);
+	else
+		fprintf(fp, "%s +%02d%02d", out,
+		            intime->offset / 60, intime->offset % 60);
 }
 
 void
 printtimeshort(FILE *fp, const git_time *intime)
 {
-	printtimeformat(fp, intime, "%Y-%m-%d %H:%M");
+	struct tm *intm;
+	time_t t;
+	char out[32];
+
+	t = (time_t)intime->time;
+	if (!(intm = gmtime(&t)))
+		return;
+	strftime(out, sizeof(out), "%Y-%m-%d %H:%M", intm);
+	fputs(out, fp);
 }
 
-int
+void
 writeheader(FILE *fp, const char *title)
 {
 	fputs("<!DOCTYPE html>\n"
@@ -242,9 +333,9 @@ writeheader(FILE *fp, const char *title)
 		"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n"
 		"<meta http-equiv=\"Content-Language\" content=\"en\" />\n<title>", fp);
 	xmlencode(fp, title, strlen(title));
-	if (title[0] && stripped_name[0])
+	if (title[0] && strippedname[0])
 		fputs(" - ", fp);
-	xmlencode(fp, stripped_name, strlen(stripped_name));
+	xmlencode(fp, strippedname, strlen(strippedname));
 	if (description[0])
 		fputs(" - ", fp);
 	xmlencode(fp, description, strlen(description));
@@ -256,7 +347,7 @@ writeheader(FILE *fp, const char *title)
 	fprintf(fp, "<a href=\"../%s\"><img src=\"%slogo.png\" alt=\"\" width=\"32\" height=\"32\" /></a>",
 	        relpath, relpath);
 	fputs("</td><td><h1>", fp);
-	xmlencode(fp, stripped_name, strlen(stripped_name));
+	xmlencode(fp, strippedname, strlen(strippedname));
 	fputs("</h1><span class=\"desc\">", fp);
 	xmlencode(fp, description, strlen(description));
 	fputs("</span></td></tr>", fp);
@@ -271,19 +362,19 @@ writeheader(FILE *fp, const char *title)
 	fprintf(fp, "<a href=\"%slog.html\">Log</a> | ", relpath);
 	fprintf(fp, "<a href=\"%sfiles.html\">Files</a> | ", relpath);
 	fprintf(fp, "<a href=\"%srefs.html\">Refs</a>", relpath);
+	if (hassubmodules)
+		fprintf(fp, " | <a href=\"%sfile/.gitmodules.html\">Submodules</a>", relpath);
 	if (hasreadme)
 		fprintf(fp, " | <a href=\"%sfile/README.html\">README</a>", relpath);
 	if (haslicense)
 		fprintf(fp, " | <a href=\"%sfile/LICENSE.html\">LICENSE</a>", relpath);
 	fputs("</td></tr></table>\n<hr/>\n<div id=\"content\">\n", fp);
-
-	return 0;
 }
 
-int
+void
 writefooter(FILE *fp)
 {
-	return !fputs("</div>\n</body>\n</html>\n", fp);
+	fputs("</div>\n</body>\n</html>\n", fp);
 }
 
 int
@@ -350,42 +441,70 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 	const git_diff_hunk *hunk;
 	const git_diff_line *line;
 	git_patch *patch;
-	git_buf statsbuf;
-	size_t ndeltas, nhunks, nhunklines;
-	size_t i, j, k;
+	size_t nhunks, nhunklines, changed, add, del, total, i, j, k;
+	char linestr[80];
 
 	printcommit(fp, ci);
 
-	memset(&statsbuf, 0, sizeof(statsbuf));
+	if (!ci->deltas)
+		return;
+
+	if (ci->filecount > 1000   ||
+	    ci->ndeltas   > 1000   ||
+	    ci->addcount  > 100000 ||
+	    ci->delcount  > 100000) {
+		fputs("Diff is too large, output suppressed.\n", fp);
+		return;
+	}
 
 	/* diff stat */
-	if (ci->stats &&
-	    !git_diff_stats_to_buf(&statsbuf, ci->stats,
-                                   GIT_DIFF_STATS_FULL | GIT_DIFF_STATS_SHORT, 80)) {
-		if (statsbuf.ptr && statsbuf.ptr[0]) {
-			fputs("<b>Diffstat:</b>\n", fp);
-			xmlencode(fp, statsbuf.ptr, strlen(statsbuf.ptr));
+	fputs("<b>Diffstat:</b>\n<table>", fp);
+	for (i = 0; i < ci->ndeltas; i++) {
+		delta = git_patch_get_delta(ci->deltas[i]->patch);
+		fprintf(fp, "<tr><td><a href=\"#h%zu\">", i);
+		xmlencode(fp, delta->old_file.path, strlen(delta->old_file.path));
+		if (strcmp(delta->old_file.path, delta->new_file.path)) {
+			fputs(" -&gt; ", fp);
+			xmlencode(fp, delta->new_file.path, strlen(delta->new_file.path));
 		}
+
+		add = ci->deltas[i]->addcount;
+		del = ci->deltas[i]->delcount;
+		changed = add + del;
+		total = sizeof(linestr) - 2;
+		if (changed > total) {
+			if (add)
+				add = ((float)total / changed * add) + 1;
+			if (del)
+				del = ((float)total / changed * del) + 1;
+		}
+		memset(&linestr, '+', add);
+		memset(&linestr[add], '-', del);
+
+		fprintf(fp, "</a></td><td> | </td><td class=\"num\">%zu</td><td><span class=\"i\">",
+		        ci->deltas[i]->addcount + ci->deltas[i]->delcount);
+		fwrite(&linestr, 1, add, fp);
+		fputs("</span><span class=\"d\">", fp);
+		fwrite(&linestr[add], 1, del, fp);
+		fputs("</span></td></tr>\n", fp);
 	}
+	fprintf(fp, "</table>%zu file%s changed, %zu insertion%s(+), %zu deletion%s(-)\n",
+		ci->filecount, ci->filecount == 1 ? "" : "s",
+	        ci->addcount,  ci->addcount  == 1 ? "" : "s",
+	        ci->delcount,  ci->delcount  == 1 ? "" : "s");
 
 	fputs("<hr/>", fp);
 
-	ndeltas = git_diff_num_deltas(ci->diff);
-	for (i = 0; i < ndeltas; i++) {
-		if (git_patch_from_diff(&patch, ci->diff, i)) {
-			git_patch_free(patch);
-			break;
-		}
-
+	for (i = 0; i < ci->ndeltas; i++) {
+		patch = ci->deltas[i]->patch;
 		delta = git_patch_get_delta(patch);
-		fprintf(fp, "<b>diff --git a/<a href=\"%sfile/%s.html\">%s</a> b/<a href=\"%sfile/%s.html\">%s</a></b>\n",
-			relpath, delta->old_file.path, delta->old_file.path,
+		fprintf(fp, "<b>diff --git a/<a id=\"h%zu\" href=\"%sfile/%s.html\">%s</a> b/<a href=\"%sfile/%s.html\">%s</a></b>\n",
+			i, relpath, delta->old_file.path, delta->old_file.path,
 			relpath, delta->new_file.path, delta->new_file.path);
 
 		/* check binary data */
 		if (delta->flags & GIT_DIFF_FLAG_BINARY) {
-			fputs("Binary files differ\n", fp);
-			git_patch_free(patch);
+			fputs("Binary files differ.\n", fp);
 			continue;
 		}
 
@@ -394,7 +513,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 			if (git_patch_get_hunk(&hunk, &nhunklines, patch, j))
 				break;
 
-			fprintf(fp, "<a href=\"#h%zu\" id=\"h%zu\" class=\"h\">", j, j);
+			fprintf(fp, "<a href=\"#h%zu-%zu\" id=\"h%zu-%zu\" class=\"h\">", i, j, i, j);
 			xmlencode(fp, hunk->header, hunk->header_len);
 			fputs("</a>", fp);
 
@@ -402,11 +521,11 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 				if (git_patch_get_line_in_hunk(&line, patch, j, k))
 					break;
 				if (line->old_lineno == -1)
-					fprintf(fp, "<a href=\"#h%zu-%zu\" id=\"h%zu-%zu\" class=\"i\">+",
-						j, k, j, k);
+					fprintf(fp, "<a href=\"#h%zu-%zu-%zu\" id=\"h%zu-%zu-%zu\" class=\"i\">+",
+						i, j, k, i, j, k);
 				else if (line->new_lineno == -1)
-					fprintf(fp, "<a href=\"#h%zu-%zu\" id=\"h%zu-%zu\" class=\"d\">-",
-						j, k, j, k);
+					fprintf(fp, "<a href=\"#h%zu-%zu-%zu\" id=\"h%zu-%zu-%zu\" class=\"d\">-",
+						i, j, k, i, j, k);
 				else
 					fputc(' ', fp);
 				xmlencode(fp, line->content, line->content_len);
@@ -414,11 +533,38 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 					fputs("</a>", fp);
 			}
 		}
-		git_patch_free(patch);
 	}
-	git_buf_free(&statsbuf);
+}
 
-	return;
+void
+writelogline(FILE *fp, struct commitinfo *ci)
+{
+	size_t len;
+
+	fputs("<tr><td>", fp);
+	if (ci->author)
+		printtimeshort(fp, &(ci->author->when));
+	fputs("</td><td>", fp);
+	if (ci->summary) {
+		fprintf(fp, "<a href=\"%scommit/%s.html\">", relpath, ci->oid);
+		if ((len = strlen(ci->summary)) > summarylen) {
+			xmlencode(fp, ci->summary, summarylen - 1);
+			fputs("…", fp);
+		} else {
+			xmlencode(fp, ci->summary, len);
+		}
+		fputs("</a>", fp);
+	}
+	fputs("</td><td>", fp);
+	if (ci->author)
+		xmlencode(fp, ci->author->name, strlen(ci->author->name));
+	fputs("</td><td class=\"num\">", fp);
+	fprintf(fp, "%zu", ci->filecount);
+	fputs("</td><td class=\"num\">", fp);
+	fprintf(fp, "+%zu", ci->addcount);
+	fputs("</td><td class=\"num\">", fp);
+	fprintf(fp, "-%zu", ci->delcount);
+	fputs("</td></tr>\n", fp);
 }
 
 int
@@ -427,7 +573,6 @@ writelog(FILE *fp, const git_oid *oid)
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
-	size_t len;
 	char path[PATH_MAX];
 	FILE *fpfile;
 	int r;
@@ -437,40 +582,17 @@ writelog(FILE *fp, const git_oid *oid)
 	git_revwalk_sorting(w, GIT_SORT_TIME);
 	git_revwalk_simplify_first_parent(w);
 
-	fputs("<table id=\"log\"><thead>\n<tr><td>Date</td><td>Commit message</td>"
-		  "<td>Author</td><td class=\"num\">Files</td><td class=\"num\">+</td>"
-		  "<td class=\"num\">-</td></tr>\n</thead><tbody>\n", fp);
-
 	while (!git_revwalk_next(&id, w)) {
 		relpath = "";
 
+		if (cachefile && !memcmp(&id, &lastoid, sizeof(id)))
+			break;
 		if (!(ci = commitinfo_getbyoid(&id)))
 			break;
 
-		fputs("<tr><td>", fp);
-		if (ci->author)
-			printtimeshort(fp, &(ci->author->when));
-		fputs("</td><td>", fp);
-		if (ci->summary) {
-			fprintf(fp, "<a href=\"%scommit/%s.html\">", relpath, ci->oid);
-			if ((len = strlen(ci->summary)) > summarylen) {
-				xmlencode(fp, ci->summary, summarylen - 1);
-				fputs("…", fp);
-			} else {
-				xmlencode(fp, ci->summary, len);
-			}
-			fputs("</a>", fp);
-		}
-		fputs("</td><td>", fp);
-		if (ci->author)
-			xmlencode(fp, ci->author->name, strlen(ci->author->name));
-		fputs("</td><td class=\"num\">", fp);
-		fprintf(fp, "%zu", ci->filecount);
-		fputs("</td><td class=\"num\">", fp);
-		fprintf(fp, "+%zu", ci->addcount);
-		fputs("</td><td class=\"num\">", fp);
-		fprintf(fp, "-%zu", ci->delcount);
-		fputs("</td></tr>\n", fp);
+		writelogline(fp, ci);
+		if (cachefile)
+			writelogline(wcachefp, ci);
 
 		relpath = "../";
 
@@ -490,8 +612,6 @@ writelog(FILE *fp, const git_oid *oid)
 		}
 		commitinfo_free(ci);
 	}
-	fputs("</tbody></table>", fp);
-
 	git_revwalk_free(w);
 
 	relpath = "";
@@ -506,8 +626,13 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 
 	fprintf(fp, "<id>%s</id>\n", ci->oid);
 	if (ci->author) {
-		fputs("<updated>", fp);
+		fputs("<published>", fp);
 		printtimez(fp, &(ci->author->when));
+		fputs("</published>\n", fp);
+	}
+	if (ci->committer) {
+		fputs("<updated>", fp);
+		printtimez(fp, &(ci->committer->when));
 		fputs("</updated>\n", fp);
 	}
 	if (ci->summary) {
@@ -543,9 +668,7 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 		fputc('\n', fp);
 		xmlencode(fp, ci->msg, strlen(ci->msg));
 	}
-	fputs("\n</content>\n", fp);
-
-	fputs("</entry>\n", fp);
+	fputs("\n</content>\n</entry>\n", fp);
 }
 
 int
@@ -554,11 +677,11 @@ writeatom(FILE *fp)
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
-	size_t i, m = 100; /* max */
+	size_t i, m = 100; /* last 'm' commits */
 
 	fputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
 	      "<feed xmlns=\"http://www.w3.org/2005/Atom\">\n<title>", fp);
-	xmlencode(fp, stripped_name, strlen(stripped_name));
+	xmlencode(fp, strippedname, strlen(strippedname));
 	fputs(", branch HEAD</title>\n<subtitle>", fp);
 	xmlencode(fp, description, strlen(description));
 	fputs("</subtitle>\n", fp);
@@ -584,24 +707,21 @@ writeatom(FILE *fp)
 int
 writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t filesize)
 {
-	char tmp[PATH_MAX] = "";
-	char *d;
+	char tmp[PATH_MAX] = "", *d;
 	const char *p;
 	int lc = 0;
 	FILE *fp;
 
-	d = xdirname(fpath);
-	if (mkdirp(d)) {
-		free(d);
+	if (strlcpy(tmp, fpath, sizeof(tmp)) >= sizeof(tmp))
+		errx(1, "path truncated: '%s'", fpath);
+	if (!(d = dirname(tmp)))
+		err(1, "dirname");
+	if (mkdirp(d))
 		return -1;
-	}
-	free(d);
 
-	p = fpath;
-	while (*p) {
+	for (p = fpath, tmp[0] = '\0'; *p; p++) {
 		if (*p == '/' && strlcat(tmp, "../", sizeof(tmp)) >= sizeof(tmp))
 			errx(1, "path truncated: '../%s'", tmp);
-		p++;
 	}
 	relpath = tmp;
 
@@ -613,7 +733,7 @@ writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t fi
 	fputs("</p><hr/>", fp);
 
 	if (git_blob_is_binary((git_blob *)obj)) {
-		fputs("<p>Binary file</p>\n", fp);
+		fputs("<p>Binary file.</p>\n", fp);
 	} else {
 		lc = writeblobhtml(fp, (git_blob *)obj);
 		if (ferror(fp))
@@ -673,60 +793,63 @@ int
 writefilestree(FILE *fp, git_tree *tree, const char *branch, const char *path)
 {
 	const git_tree_entry *entry = NULL;
-	const char *entryname;
-	char filepath[PATH_MAX], entrypath[PATH_MAX];
+	git_submodule *module = NULL;
 	git_object *obj = NULL;
 	git_off_t filesize;
+	const char *entryname;
+	char filepath[PATH_MAX], entrypath[PATH_MAX];
 	size_t count, i;
 	int lc, r, ret;
 
 	count = git_tree_entrycount(tree);
 	for (i = 0; i < count; i++) {
 		if (!(entry = git_tree_entry_byindex(tree, i)) ||
-		    git_tree_entry_to_object(&obj, repo, entry))
+		    !(entryname = git_tree_entry_name(entry)))
 			return -1;
-		entryname = git_tree_entry_name(entry);
-		r = snprintf(entrypath, sizeof(entrypath), "%s%s%s",
-			 path, path[0] ? "/" : "", entryname);
-		if (r == -1 || (size_t)r >= sizeof(entrypath))
-			errx(1, "path truncated: '%s%s%s'",
-			        path, path[0] ? "/" : "", entryname);
-		switch (git_object_type(obj)) {
-		case GIT_OBJ_BLOB:
-			break;
-		case GIT_OBJ_TREE:
-			/* NOTE: recurses */
-			ret = writefilestree(fp, (git_tree *)obj, branch,
-			                     entrypath);
-			git_object_free(obj);
-			if (ret)
-				return ret;
-			continue;
-		default:
-			git_object_free(obj);
-			continue;
-		}
+		joinpath(entrypath, sizeof(entrypath), path, entryname);
 
-		r = snprintf(filepath, sizeof(filepath), "file/%s%s%s.html",
-		         path, path[0] ? "/" : "", entryname);
+		r = snprintf(filepath, sizeof(filepath), "file/%s.html",
+		         entrypath);
 		if (r == -1 || (size_t)r >= sizeof(filepath))
-			errx(1, "path truncated: 'file/%s%s%s.html'",
-			        path, path[0] ? "/" : "", entryname);
+			errx(1, "path truncated: 'file/%s.html'", entrypath);
 
-		filesize = git_blob_rawsize((git_blob *)obj);
+		if (!git_tree_entry_to_object(&obj, repo, entry)) {
+			switch (git_object_type(obj)) {
+			case GIT_OBJ_BLOB:
+				break;
+			case GIT_OBJ_TREE:
+				/* NOTE: recurses */
+				ret = writefilestree(fp, (git_tree *)obj, branch,
+				                     entrypath);
+				git_object_free(obj);
+				if (ret)
+					return ret;
+				continue;
+			default:
+				git_object_free(obj);
+				continue;
+			}
 
-		lc = writeblob(obj, filepath, entryname, filesize);
+			filesize = git_blob_rawsize((git_blob *)obj);
+			lc = writeblob(obj, filepath, entryname, filesize);
 
-		fputs("<tr><td>", fp);
-		fputs(filemode(git_tree_entry_filemode(entry)), fp);
-		fprintf(fp, "</td><td><a href=\"%s%s\">", relpath, filepath);
-		xmlencode(fp, entrypath, strlen(entrypath));
-		fputs("</a></td><td class=\"num\">", fp);
-		if (showlinecount && lc > 0)
-			fprintf(fp, "%dL", lc);
-		else
-			fprintf(fp, "%juB", (uintmax_t)filesize);
-		fputs("</td></tr>\n", fp);
+			fputs("<tr><td>", fp);
+			fputs(filemode(git_tree_entry_filemode(entry)), fp);
+			fprintf(fp, "</td><td><a href=\"%s%s\">", relpath, filepath);
+			xmlencode(fp, entrypath, strlen(entrypath));
+			fputs("</a></td><td class=\"num\">", fp);
+			if (showlinecount && lc > 0)
+				fprintf(fp, "%dL", lc);
+			else
+				fprintf(fp, "%juB", (uintmax_t)filesize);
+			fputs("</td></tr>\n", fp);
+		} else if (!git_submodule_lookup(&module, repo, entryname)) {
+			fprintf(fp, "<tr><td>m---------</td><td><a href=\"%sfile/.gitmodules.html\">",
+				relpath);
+			xmlencode(fp, entrypath, strlen(entrypath));
+			git_submodule_free(module);
+			fputs("</a></td><td class=\"num\"></td></tr>\n", fp);
+		}
 	}
 
 	return 0;
@@ -743,12 +866,10 @@ writefiles(FILE *fp, const git_oid *id, const char *branch)
 	      "<td>Mode</td><td>Name</td><td class=\"num\">Size</td>"
 	      "</tr>\n</thead><tbody>\n", fp);
 
-	if (git_commit_lookup(&commit, repo, id) ||
-	    git_commit_tree(&tree, commit))
-		goto err;
-	ret = writefilestree(fp, tree, branch, "");
+	if (!git_commit_lookup(&commit, repo, id) &&
+	    !git_commit_tree(&tree, commit))
+		ret = writefilestree(fp, tree, branch, "");
 
-err:
 	fputs("</tbody></table>", fp);
 
 	git_commit_free(commit);
@@ -783,7 +904,7 @@ writerefs(FILE *fp)
 	git_reference *dref = NULL, *r, *ref = NULL;
 	git_reference_iterator *it = NULL;
 	git_reference **refs = NULL;
-	size_t count, i, j, refcount = 0;
+	size_t count, i, j, refcount;
 	const char *titles[] = { "Branches", "Tags" };
 	const char *ids[] = { "branches", "tags" };
 	const char *name;
@@ -819,9 +940,8 @@ writerefs(FILE *fp)
 			default:
 				continue;
 			}
-			if (!(id = git_reference_target(r)))
-				goto err;
-			if (git_reference_peel(&obj, r, GIT_OBJ_ANY))
+			if (!git_reference_target(r) ||
+			    git_reference_peel(&obj, r, GIT_OBJ_ANY))
 				goto err;
 			if (!(id = git_object_id(obj)))
 				goto err;
@@ -872,6 +992,13 @@ err:
 	return 0;
 }
 
+void
+usage(char *argv0)
+{
+	fprintf(stderr, "%s [-c cachefile] repodir\n", argv0);
+	exit(1);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -879,22 +1006,38 @@ main(int argc, char *argv[])
 	const git_oid *head = NULL;
 	const git_error *e = NULL;
 	FILE *fp, *fpread;
-	char path[PATH_MAX], *p;
-	int r, status;
+	char path[PATH_MAX], repodirabs[PATH_MAX + 1], *p;
+	char tmppath[64] = "cache.XXXXXXXXXXXX", buf[BUFSIZ];
+	size_t n;
+	int i, fd;
 
-	if (argc != 2) {
-		fprintf(stderr, "%s <repodir>\n", argv[0]);
-		return 1;
+	if (pledge("stdio rpath wpath cpath", NULL) == -1)
+		err(1, "pledge");
+
+	for (i = 1; i < argc; i++) {
+		if (argv[i][0] != '-') {
+			if (repodir)
+				usage(argv[0]);
+			repodir = argv[i];
+		} else if (argv[i][1] == 'c') {
+			if (i + 1 >= argc)
+				usage(argv[0]);
+			cachefile = argv[++i];
+		}
 	}
-	repodir = argv[1];
+	if (!repodir)
+		usage(argv[0]);
+
+	if (!realpath(repodir, repodirabs))
+		err(1, "realpath");
 
 	git_libgit2_init();
 
-	if ((status = git_repository_open_ext(&repo, repodir,
-		GIT_REPOSITORY_OPEN_NO_SEARCH, NULL)) < 0) {
+	if (git_repository_open_ext(&repo, repodir,
+		GIT_REPOSITORY_OPEN_NO_SEARCH, NULL) < 0) {
 		e = giterr_last();
-		fprintf(stderr, "error %d/%d: %s\n", status, e->klass, e->message);
-		return status;
+		fprintf(stderr, "%s: %s\n", argv[0], e->message);
+		return 1;
 	}
 
 	/* find HEAD */
@@ -904,27 +1047,22 @@ main(int argc, char *argv[])
 	git_object_free(obj);
 
 	/* use directory name as name */
-	name = xbasename(repodir);
+	if ((name = strrchr(repodirabs, '/')))
+		name++;
+	else
+		name = "";
 
 	/* strip .git suffix */
-	if (!(stripped_name = strdup(name)))
+	if (!(strippedname = strdup(name)))
 		err(1, "strdup");
-	if ((p = strrchr(stripped_name, '.')))
+	if ((p = strrchr(strippedname, '.')))
 		if (!strcmp(p, ".git"))
 			*p = '\0';
 
 	/* read description or .git/description */
-	r = snprintf(path, sizeof(path), "%s%s%s",
-		repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
-	if (r == -1 || (size_t)r >= sizeof(path))
-		errx(1, "path truncated: '%s%s%s'",
-	                repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
+	joinpath(path, sizeof(path), repodir, "description");
 	if (!(fpread = fopen(path, "r"))) {
-		r = snprintf(path, sizeof(path), "%s%s%s",
-			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: '%s%s%s'",
-		                repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
+		joinpath(path, sizeof(path), repodir, ".git/description");
 		fpread = fopen(path, "r");
 	}
 	if (fpread) {
@@ -934,17 +1072,9 @@ main(int argc, char *argv[])
 	}
 
 	/* read url or .git/url */
-	r = snprintf(path, sizeof(path), "%s%s%s",
-		repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "url");
-	if (r == -1 || (size_t)r >= sizeof(path))
-		errx(1, "path truncated: '%s%s%s'",
-		        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "url");
+	joinpath(path, sizeof(path), repodir, "url");
 	if (!(fpread = fopen(path, "r"))) {
-		r = snprintf(path, sizeof(path), "%s%s%s",
-			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/url");
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: '%s%s%s'",
-			        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/url");
+		joinpath(path, sizeof(path), repodir, ".git/url");
 		fpread = fopen(path, "r");
 	}
 	if (fpread) {
@@ -960,13 +1090,56 @@ main(int argc, char *argv[])
 	/* check README */
 	hasreadme = !git_revparse_single(&obj, repo, "HEAD:README");
 	git_object_free(obj);
+	hassubmodules = !git_revparse_single(&obj, repo, "HEAD:.gitmodules");
+	git_object_free(obj);
 
 	/* log for HEAD */
 	fp = efopen("log.html", "w");
 	relpath = "";
-	writeheader(fp, "Log");
 	mkdir("commit", 0755);
-	writelog(fp, head);
+	writeheader(fp, "Log");
+	fputs("<table id=\"log\"><thead>\n<tr><td>Date</td><td>Commit message</td>"
+		  "<td>Author</td><td class=\"num\">Files</td><td class=\"num\">+</td>"
+		  "<td class=\"num\">-</td></tr>\n</thead><tbody>\n", fp);
+
+	if (cachefile) {
+		/* read from cache file (does not need to exist) */
+		if ((rcachefp = fopen(cachefile, "r"))) {
+			if (!fgets(lastoidstr, sizeof(lastoidstr), rcachefp))
+				errx(1, "%s: no object id", cachefile);
+			if (git_oid_fromstr(&lastoid, lastoidstr))
+				errx(1, "%s: invalid object id", cachefile);
+		}
+
+		/* write log to (temporary) cache */
+		if ((fd = mkstemp(tmppath)) == -1)
+			err(1, "mkstemp");
+		if (!(wcachefp = fdopen(fd, "w")))
+			err(1, "fdopen");
+		/* write last commit id (HEAD) */
+		git_oid_tostr(buf, sizeof(buf), head);
+		fprintf(wcachefp, "%s\n", buf);
+
+		writelog(fp, head);
+
+		if (rcachefp) {
+			/* append previous log to log.html and the new cache */
+			while (!feof(rcachefp)) {
+				n = fread(buf, 1, sizeof(buf), rcachefp);
+				if (ferror(rcachefp))
+					err(1, "fread");
+				if (fwrite(buf, 1, n, fp) != n ||
+				    fwrite(buf, 1, n, wcachefp) != n)
+					err(1, "fwrite");
+			}
+			fclose(rcachefp);
+		}
+		fclose(wcachefp);
+	} else {
+		writelog(fp, head);
+	}
+
+	fputs("</tbody></table>", fp);
 	writefooter(fp);
 	fclose(fp);
 
@@ -989,6 +1162,10 @@ main(int argc, char *argv[])
 	writeatom(fp);
 	fclose(fp);
 
+	/* rename new cache file on success */
+	if (cachefile && rename(tmppath, cachefile))
+		err(1, "rename: '%s' to '%s'", tmppath, cachefile);
+
 	/* cleanup */
 	git_repository_free(repo);
 	git_libgit2_shutdown();
